@@ -19,6 +19,7 @@ __metaclass__ = type
 
 import socket
 import sys
+import time
 
 from paste.request import construct_url
 
@@ -46,9 +47,13 @@ default_map_environ = {
     }
 
 
+class SoftRequestTimeout(Exception):
+    """Soft request timeout expired"""
+
+
 def make_app(app, config, template=default_error_template,
         content_type='text/html', error_render=None, oops_on_status=None,
-        map_environ=None, tracker=None):
+        map_environ=None, tracker=None, soft_start_timeout=None):
     """Construct a middleware around app that will forward errors via config.
 
     Any errors encountered by the app will be forwarded to config and an error
@@ -86,6 +91,13 @@ def make_app(app, config, template=default_error_template,
         See generator_tracker for the reference tracker used in regular WSGI
         environments. generator_tracker is used by default or when
         tracker=None.
+    :param soft_start_timeout: A duration in milliseconds for the creation of
+        reports on slow requests. If this is set and the duration between
+        calling into the app and start_response being called is greater than
+        the timeout value, then an OOPS will be created and the OOPS id added
+        to the response HTTP headers as normal. A backtrace leading into the
+        middleware is generated (this can be informative as start_response is
+        a callback) and the exception type is set to SoftRequestTimeout.
     :return: A WSGI app.
     """
     def oops_middleware(environ, start_response):
@@ -100,6 +112,8 @@ def make_app(app, config, template=default_error_template,
         """
         environ['oops.report'] = {}
         environ['oops.context'] = {}
+        if soft_start_timeout:
+            start_time = time.time()
         state = {}
         def make_context(exc_info=None):
             context = dict(url=construct_url(environ), wsgi_environ=environ)
@@ -125,9 +139,11 @@ def make_app(app, config, template=default_error_template,
             if exc_info is not None:
                 # The app is explicitly signalling an error (rather than
                 # returning a page describing the error). Capture that and then
-                # forward to the containing element verbatim. In future we might
-                # choose to add the OOPS id to the headers (but we can't fiddle
-                # the body as it is being generated by the contained app.
+                # forward to the containing element untouched except for the
+                # addition of the X-Oops-Id header. We don't touch the body
+                # because the application is handling the error and generating
+                # the body itself. We may in future provide an option to
+                # replace the body in this situation.
                 report = config.create(make_context(exc_info=exc_info))
                 ids = config.publish(report)
                 try:
@@ -139,12 +155,24 @@ def make_app(app, config, template=default_error_template,
                 finally:
                     del exc_info
             else:
+                do_oops = False
                 if oops_on_status:
                     for sniff_status in oops_on_status:
                         if status.startswith(sniff_status):
-                            report = config.create(make_context())
-                            report['HTTP_STATUS'] = status.split(' ')[0]
-                            config.publish(report)
+                            do_oops = True
+                if (soft_start_timeout and
+                    (time.time()-start_time)*1000 > soft_start_timeout):
+                    try:
+                        raise SoftRequestTimeout(
+                            "Start_response over timeout %s."
+                            % soft_start_timeout)
+                    except SoftRequestTimeout:
+                        exc_info = sys.exc_info()
+                    do_oops = True
+                if do_oops:
+                    report = config.create(make_context(exc_info=exc_info))
+                    report['HTTP_STATUS'] = status.split(' ')[0]
+                    config.publish(report)
                 state['response'] = (status, headers)
             return oops_write
         try:
